using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;

namespace ModelEx3.AMF
{
    public partial class FileSaveList : Form
    {
        public interface Entry
        {
            String FileName { get; }
            Object Object { get; }
            bool Save { get; set; }
            bool Enabled { get; set;  }
        }

        public class InternalEntry : ListViewExItem, Entry
        {
            public String FileName { get { return SubItems[1].Text; } }
            public Object Object { get { return Tag; } }
            public bool Save
            {
                get { return Checked; }
                set { Checked = value; }
            }

            public InternalEntry(String strFileName, Object xObject)
            {
                UseItemStyleForSubItems = true;
                Checked = true;
                Tag = xObject;
                SubItems.Add(strFileName);
                IndentCount = 1;
            }
        }

        protected List<Entry> m_xEntries;

        public String FolderName
        {
            get { return comboBox1.Text; }
            set
            {
                comboBox1.Items.Clear();
                comboBox1.Items.Insert(0, value);
                comboBox1.Text = value;
            }
        }

        public Entry[] Entries { get { return m_xEntries.ToArray();  } }

        public bool IsEmpty { get { return listView1.Items.Count == 0; } }

        public FileSaveList()
        {
            InitializeComponent();
            m_xEntries = new List<Entry>();

            listView1.ItemAdded += new ListViewItemAddEventHandler(listView1_ItemAdd);
            listView1.ItemEnabled += new ListViewItemEnableEventHandler(listView1_ItemEnable);

            listView1.SmallImageList = new ImageList();
            Icon GreenSphere = new Icon(GetType().Assembly.GetManifestResourceStream("ModelEx3.small-lime-sphere.ico"));
            listView1.SmallImageList.Images.Add(GreenSphere);
            Icon YellowSphere = new Icon(GetType().Assembly.GetManifestResourceStream("ModelEx3.small-yellow-sphere.ico"));
            listView1.SmallImageList.Images.Add(YellowSphere);
            Icon WhiteSphere = new Icon(GetType().Assembly.GetManifestResourceStream("ModelEx3.small-white-sphere.ico"));
            listView1.SmallImageList.Images.Add(WhiteSphere);
        }

        void listView1_ItemAdd(object sender, ListViewItemAddEventArgs e)
        {
            int iItemIndex = e.Item.Index;
            InternalEntry xEntry = (InternalEntry)e.Item;

            if (!xEntry.Enabled)
            {
                xEntry.ForeColor = Color.LightGray;
                listView1.SetSubItemImageIndex(iItemIndex, 1, 2);
            }
            else
            {
                xEntry.ForeColor = Color.Black;
                if (File.Exists(xEntry.FileName))
                {
                    listView1.SetSubItemImageIndex(iItemIndex, 1, 1);
                }
                else
                {
                    listView1.SetSubItemImageIndex(iItemIndex, 1, 0);
                }
            }
        }

        void listView1_ItemEnable(object sender, ListViewItemEnableEventArgs e)
        {
            int iItemIndex = e.Item.Index;
            InternalEntry xEntry = (InternalEntry)e.Item;

            if (!xEntry.Enabled)
            {
                xEntry.ForeColor = Color.LightGray;
                listView1.SetSubItemImageIndex(iItemIndex, 1, 2);
            }
            else
            {
                xEntry.ForeColor = Color.Black;
                if (File.Exists(xEntry.FileName))
                {
                    listView1.SetSubItemImageIndex(iItemIndex, 1, 1);
                }
                else
                {
                    listView1.SetSubItemImageIndex(iItemIndex, 1, 0);
                }
            }
        }

        private void okButton_Click(object sender, EventArgs e)
        {
            DialogResult = DialogResult.OK;
        }

        private void cancelButton_Click(object sender, EventArgs e)
        {
            DialogResult = DialogResult.Cancel;
        }

        private void browseButton_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog xFolderDlg = new FolderBrowserDialog();
            xFolderDlg.Description = "Save all files to the current folder";
            xFolderDlg.ShowNewFolderButton = true;
            xFolderDlg.SelectedPath = Directory.GetCurrentDirectory();
            if (xFolderDlg.ShowDialog() == DialogResult.OK)
            {
                while (comboBox1.Items.Contains(xFolderDlg.SelectedPath))
                {
                    comboBox1.Items.Remove(xFolderDlg.SelectedPath);
                }
                comboBox1.Items.Insert(0, xFolderDlg.SelectedPath);
                comboBox1.Text = xFolderDlg.SelectedPath;
            }
        }

        public void Add(String strFileName, Object xObject)
        {
            if (listView1.Items.Count > 0 &&
                listView1.FindItemWithText(strFileName, true, 0) != null)
            {
                return;
            }

            InternalEntry xEntry = new InternalEntry(strFileName, xObject);
            xEntry.Enabled = ValidateObject(xObject);
            listView1.Items.Add(xEntry);
            m_xEntries.Add(xEntry);
        }

        public void Remove(String strFileName)
        {
            while (listView1.Items.Count > 0)
            {
                ListViewItem xItem = listView1.FindItemWithText(strFileName, true, 0);
                if (xItem != null)
                {
                    listView1.Items.Remove(xItem);
                }
                else
                {
                    break;
                }
            }
        }

        public void Remove(String[] astrFileNames)
        {
            foreach (String strFileName in astrFileNames)
            {
                Remove(strFileName);
            }
        }

        public void Clear()
        {
            m_xEntries.Clear();
            listView1.Items.Clear();
        }

        protected virtual bool ValidateObject(Object xObject)
        {
            return true;
        }
    }
}