using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace ModelEx3
{
    public delegate void ListViewItemAddEventHandler(object sender, ListViewItemAddEventArgs e);

    public class ListViewItemAddEventArgs : EventArgs
    {
        protected ListViewItem item;

        public ListViewItem Item { get { return item; } }

        public ListViewItemAddEventArgs(ListViewItem item)
        {
            this.item = item;
        }
    }

    public delegate void ListViewItemEnableEventHandler(object sender, ListViewItemEnableEventArgs e);

    public class ListViewItemEnableEventArgs : EventArgs
    {
        protected ListViewItem item;

        public ListViewItem Item { get { return item; } }

        public ListViewItemEnableEventArgs(ListViewItem item)
        {
            this.item = item;
        }
    }

    public class ListViewExItem : ListViewItem
    {
        protected bool m_bEnabled = true;

        public bool Enabled
        {
            get { return m_bEnabled; }
            set
            {
                m_bEnabled = value;
                if (ListView != null && ListView.GetType() == typeof(ListViewEx))
                {
                    ListViewEx xListView = (ListViewEx)ListView;
                    ListViewItemEnableEventArgs e = new ListViewItemEnableEventArgs(this);
                    if (xListView.ItemEnabled != null)
                    {
                        xListView.ItemEnabled(xListView, e);
                    }
                }
            }
        }
    }

    public class ListViewEx : ListView
    {
        #region DLL Import

        private const int WM_SIZE           = 0x00000005;
        private const int WM_PAINT          = 0x0000000F;
        private const int WM_SETCURSOR      = 0x00000020;
        private const int WM_NOTIFY         = 0x0000004E;
        private const int WM_NCCALCSIZE     = 0x00000083;
        private const int WM_HSCROLL        = 0x00000114;
        private const int WM_VSCROLL        = 0x00000115;
        private const int WM_LBUTTONDBLCLK  = 0x00000203;
        private const int WM_USER           = 0x00000400;

        private const int LVIF_IMAGE = 0x2;
        private const int LVM_FIRST = 0x1000;
        private const int LVM_SETEXTENDEDLISTVIEWSTYLE = LVM_FIRST + 54;
        private const int LVS_EX_SUBITEMIMAGES = 0x2;
        private const int LVM_GETITEMA = LVM_FIRST + 5;
        private const int LVM_GETITEMW = LVM_FIRST + 75;
        private static readonly int LVM_GETITEM = (Marshal.SystemDefaultCharSize == 1) ? LVM_GETITEMA : LVM_GETITEMW;
        private const int LVM_SETITEMA = LVM_FIRST + 6;
        private const int LVM_SETITEMW = LVM_FIRST + 76;
        private static readonly int LVM_SETITEM = (Marshal.SystemDefaultCharSize == 1) ? LVM_SETITEMA : LVM_SETITEMW;
        private const int LVM_INSERTITEMA = (LVM_FIRST + 7);
        private const int LVM_INSERTITEMW = (LVM_FIRST + 77);
        private static readonly int LVM_INSERTITEM = (Marshal.SystemDefaultCharSize == 1) ? LVM_INSERTITEMA : LVM_INSERTITEMW;

        private const int OCM__BASE         = WM_USER + 0x00001C00;
        private const int OCM_NOTIFY        = OCM__BASE + WM_NOTIFY;
        private const int WS_VSCROLL        = 0x00200000;
        private const int WS_HSCROLL        = 0x00100000;
        private const int HDN_FIRST         = (0 - 300);
        private const int HDN_BEGINTRACKA   = (HDN_FIRST - 6);
        private const int HDN_BEGINTRACKW   = (HDN_FIRST - 26);
        private const int GWL_STYLE         = -16;
        private const int GWL_EXSTYLE       = -20;
        private const int SB_HORZ           = 0;
        private const int SB_VERT           = 1;
        private const int SB_CTL            = 2;
        private const int SB_BOTH           = 3;
        private const int NM_FIRST          = (0 - 0);
        private const int NM_LAST           = (0 - 99);
        private const int LVN_FIRST         = (0 - 100);
        private const int LVN_LAST          = (0 - 199);
        private const int LVN_ITEMCHANGING  = (LVN_FIRST - 0);
        private const int LVN_ITEMCHANGED   = (LVN_FIRST - 1);
        private const int LVN_INSERTITEM    = (LVN_FIRST - 2);
        private const int LVN_MARQUEEBEGIN  = (LVN_FIRST - 56);
        private const int LVIF_STATE        = 0x00000008;
        private const int LVIS_FOCUSED      = 0x00000001;
        private IntPtr    S_OK              = (IntPtr)0x00000000L;
        private IntPtr    S_FALSE           = (IntPtr)0x00000001L; 

        [StructLayout(LayoutKind.Sequential)]
        private struct RECT
        {
            public int Left;
            public int Top;
            public int Right;
            public int Bottom;
        }

        [StructLayout(LayoutKind.Sequential)]
        struct NCCALCSIZE_PARAMS
        {
            public RECT xNewCoords, xOldCoords, xOldClientCoords;
            public IntPtr lppos;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct POINT
        {
            public int x;
            public int y;
        }

        /// <summary>
        /// Notify message header structure.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        private struct NMHDR
        {
            public IntPtr hwndFrom;
            public int idFrom;
            public int code;
        }

        /// <summary>
        /// ListView notify message header structure.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        private struct NMLISTVIEW
        {
            public NMHDR hdr;
            public int iItem;
            public int iSubItem;
            public int uNewState;
            public int uOldState;
            public int uChanged;
            public POINT ptAction;
            public int lParam;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
        private struct LVITEM
        {
            public int mask;
            public int iItem;
            public int subItem;
            public int state;
            public int stateMask;
            [MarshalAs(UnmanagedType.LPTStr)]
            public string lpszText;
            public int cchTextMax;
            public int iImage;
            public IntPtr lParam;
            public int iIndent;
        }


        [DllImport("User32", CharSet = CharSet.Auto)]
        private static extern int GetWindowLong(IntPtr hWnd, int Index);

        [DllImport("User32", CharSet = CharSet.Auto)]
        private static extern int SetWindowLong(IntPtr hWnd, int Index, int Value);

        [DllImport("User32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern IntPtr SendMessage(IntPtr hWnd, int msg, IntPtr wParam, IntPtr lParam);
        [DllImport("User32.dll", CharSet = CharSet.Auto)]
        private static extern int SendMessage(IntPtr hwnd, int msg, IntPtr wParam, ref LVITEM lParam);

        [DllImport("user32.dll")]
        private static extern int ShowScrollBar(IntPtr hWnd, int wBar, bool bShow);

        #endregion

        public enum ScrollBarUse
        {
            Never       = 0,
            AsNeeded    = 1,
            Always      = 2,
        };

        private HeaderControl                           m_xHdrCtrl          = null;
        private bool                                    m_bColumnSizeLocked = true;

        private ScrollBarUse                            m_eHorizontal       = ScrollBarUse.AsNeeded;
        private ScrollBarUse                            m_eVertical         = ScrollBarUse.AsNeeded;

        private bool                                    m_bAllowItemSelect  = true;

        private event ListViewItemAddEventHandler       m_xItemAdded;
        private event ListViewItemEnableEventHandler    m_xItemEnabled;

        /// <summary>
        /// Gets or sets the ability to size the column headers.
        /// </summary>
        public bool LockColumnSize
        {
            get { return m_bColumnSizeLocked; }
            set { m_bColumnSizeLocked = value; }
        }

        /// <summary>
        /// Gets or sets the ability to select items.
        /// </summary>
        public bool AllowItemSelect
        {
            get { return m_bAllowItemSelect; }
            set { m_bAllowItemSelect = value; }
        }

        /// <summary>
        /// Gets or sets the usage for the horizontal scrollbar
        /// </summary>
        public ScrollBarUse ScrollHorizontal
        {
            get { return m_eHorizontal; }
            set { m_eHorizontal = value; }
        }

        /// <summary>
        /// Gets or sets the usage for the vertical scrollbar
        /// </summary>
        public ScrollBarUse ScrollVertical
        {
            get { return m_eVertical; }
            set { m_eVertical = value; }
        }

        public ListViewItemAddEventHandler ItemAdded
        {
            get { return m_xItemAdded; }
            set { m_xItemAdded = value; }
        }

        public ListViewItemEnableEventHandler ItemEnabled
        {
            get { return m_xItemEnabled; }
            set { m_xItemEnabled = value; }
        }

        /// <summary>
        /// Class used to capture window messages for the header of the list view
        /// control.  
        /// </summary>
        private class HeaderControl : NativeWindow
        {
            private ListViewEx parentListView = null;

            public HeaderControl(ListViewEx m)
            {
                parentListView = m;
                IntPtr header = SendMessage(
                    m.Handle, (0x1000 + 31), IntPtr.Zero, IntPtr.Zero
                );
                AssignHandle(header);
            }

            protected override void WndProc(ref Message message)
            {
                switch (message.Msg)
                {
                    case WM_LBUTTONDBLCLK:
                    case WM_SETCURSOR:
                        if (parentListView.LockColumnSize)
                        {
                            message.Result = (IntPtr)1;
                            return;
                        }
                        break;
                }

                base.WndProc(ref message);
            }
        }

        protected override void WndProc(ref Message msg)
        {
            switch (msg.Msg)
            {
                case WM_VSCROLL:
                case WM_HSCROLL:
                {
                    // Move focus to the ListView to cause ComboBox to lose focus.
                    // Focus();
                    break;
                }
                case WM_NOTIFY:
                {
                    NMHDR nmhdr = (NMHDR)msg.GetLParam(typeof(NMHDR));
                    switch (nmhdr.code)
                    {
                        case HDN_BEGINTRACKA:
                        case HDN_BEGINTRACKW:
                            if (m_bColumnSizeLocked)
                            {
                                msg.Result = (IntPtr)1;
                                return;
                            }
                            break;
                        case LVN_MARQUEEBEGIN:
                            //msg.Result = (IntPtr)1;
                            break;
                    }
                    break;
                }
                //case WM_PAINT:
                //// case WM_SIZE:
                //{
                //    if (View == View.Details && Columns.Count > 0)
                //    {
                //        Columns[Columns.Count - 1].Width = -2;
                //    }
                //    break;
                //}
                // SystemInformation
                case WM_NCCALCSIZE:
                {
                    //NCCALCSIZE_PARAMS xNCSizeParams;
                    //xNCSizeParams = (NCCALCSIZE_PARAMS)Marshal.PtrToStructure(msg.LParam, typeof(NCCALCSIZE_PARAMS));
                    //Marshal.StructureToPtr(xNCSizeParams, msg.LParam, false);

                    //if (View == View.Details && Columns.Count > 0)
                    //{
                    //    Columns[Columns.Count - 1].Width =
                    //        Width - (xNCSizeParams.xOldClientCoords.Left - xNCSizeParams.xOldCoords.Left);
                    //}

                    switch (m_eVertical)
                    {
                        case ScrollBarUse.Always:
                        {
                            ShowScrollBar(Handle, SB_VERT, true);
                            break;
                        }
                        case ScrollBarUse.Never:
                        {
                            ShowScrollBar(Handle, SB_VERT, false);
                            break;
                        }
                    }
                    switch (m_eHorizontal)
                    {
                        case ScrollBarUse.Always:
                        {
                            ShowScrollBar(Handle, SB_HORZ, true);
                            break;
                        }
                        case ScrollBarUse.Never:
                        {
                            ShowScrollBar(Handle, SB_HORZ, false);
                            break;
                        }
                    }
                    break;
                }
                case OCM_NOTIFY:
                {
                    NMHDR nmhdr = (NMHDR)msg.GetLParam(typeof(NMHDR));
                    switch (nmhdr.code)
                    {
                        case LVN_ITEMCHANGING:
                        case LVN_ITEMCHANGED:
                        {
                            NMLISTVIEW nmlv = (NMLISTVIEW)msg.GetLParam(typeof(NMLISTVIEW));

                            if ((nmlv.uChanged == LVIF_STATE) &&
                                (nmlv.uOldState & LVIS_FOCUSED) == 0 &&
                                (nmlv.uNewState & LVIS_FOCUSED) != 0)
                            {
                                switch (nmhdr.code)
                                {
                                    case LVN_ITEMCHANGING:
                                    {
                                        CancelEventArgs e = new CancelEventArgs(false);

                                        OnFocusedItemChanging(e);
                                        msg.Result = e.Cancel ? S_FALSE : S_OK;
                                        return;
                                    }
                                    case LVN_ITEMCHANGED:
                                    {
                                        OnFocusedItemChanged(System.EventArgs.Empty);
                                        break;
                                    }
                                }
                            }
                            break;
                        }
                    }
                    break;
                }
                case LVM_INSERTITEMA:
                case LVM_INSERTITEMW:
                {
                    base.WndProc(ref msg);
                    LVITEM lvi = (LVITEM)msg.GetLParam(typeof(LVITEM));
                    int itemindex = (int)msg.Result;
                    ListViewItem item = Items[itemindex];
                    ListViewItemAddEventArgs e = new ListViewItemAddEventArgs(item);
                    if (ItemAdded != null)
                    {
                        ItemAdded(this, e);
                    }
                    return;
                }
            }

            base.WndProc(ref msg);
        }

        /// <summary>
        /// The selected item is about to change.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void OnFocusedItemChanging(CancelEventArgs e)
        {
            e.Cancel = !m_bAllowItemSelect;
        }

        /// <summary>
        /// The selected item has changed.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void OnFocusedItemChanged(EventArgs e)
        {
        }

        private void ListView_SetExtendedListViewStyleEx(IntPtr hwnd, int mask, int style)
        {
            SendMessage(hwnd, LVM_SETEXTENDEDLISTVIEWSTYLE, new IntPtr(mask),
            new IntPtr(style));
        }

        private bool ListView_SetItem(IntPtr hwnd, ref LVITEM lvi)
        {
            int result = SendMessage(hwnd, LVM_SETITEM, IntPtr.Zero, ref lvi);
            return (result == 0) ? false : true;
        }

        private bool ListView_GetItem(IntPtr hwnd, ref LVITEM lvi)
        {
            int result = SendMessage(hwnd, LVM_GETITEM, IntPtr.Zero, ref lvi);
            return (result == 0) ? false : true;
        }

        private void ListView_SetSubItemImageIndex(IntPtr hwnd, int index, int subItemIndex, int imageIndex)
        {
            LVITEM lvi = new LVITEM();
            lvi.iItem = index;
            lvi.subItem = subItemIndex;
            lvi.iImage = imageIndex;
            lvi.mask = LVIF_IMAGE;
            ListView_SetItem(hwnd, ref lvi);
        }

        private int ListView_GetSubItemImageIndex(IntPtr hwnd, int index, int subItemIndex)
        {
            LVITEM lvi = new LVITEM();
            lvi.iItem = index;
            lvi.subItem = subItemIndex;
            lvi.mask = LVIF_IMAGE;
            if (ListView_GetItem(hwnd, ref lvi))
            {
                return lvi.iImage;
            }
            else
            {
                return -1;
            }
        }

        public void SetSubItemImageIndex(int index, int subItemIndex, int imageIndex)
        {
            ListView_SetSubItemImageIndex(this.Handle, index, subItemIndex, imageIndex);
        }

        public int GetSubItemImageIndex(int index, int subItemIndex)
        {
            return ListView_GetSubItemImageIndex(this.Handle, index, subItemIndex);
        }

        /// <summary>
        /// When the control is created capture the messages for the header. 
        /// </summary>
        protected override void OnCreateControl()
        {
            base.OnCreateControl();
            m_xHdrCtrl = new HeaderControl(this);
            ListView_SetExtendedListViewStyleEx(this.Handle, LVS_EX_SUBITEMIMAGES, LVS_EX_SUBITEMIMAGES);
        }

        /// <summary>
        /// Capture CTRL+ to prevent resize of all columns.
        /// </summary>
        /// <param name="e"></param>
        protected override void OnKeyDown(KeyEventArgs e)
        {
            if (e.KeyValue == (int)Keys.Add && e.Modifiers == Keys.Control && m_bColumnSizeLocked)
            {
                e.Handled = true;
            }
            else
            {
                base.OnKeyDown(e);
            }
        }

        /// <summary>
        /// Capture resizing to fit the last column to the width of the client
        /// rectangle.
        /// </summary>
        /// <param name="e"></param>
        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);
            int iBorderWidth = Bounds.Width - ClientSize.Width;
            if (View == View.Details && Columns.Count > 0)
            {
                Columns[Columns.Count - 1].Width = Width - iBorderWidth;
            }
        }
    }
}
