using System;
using System.Drawing.Imaging;
using System.Collections.Generic;
using System.Text;

namespace BenLincoln.TheLostWorlds.CDTextures
{
    public abstract class SoulReaverMonolithicTextureFile : TextureFile
    {
        // Fields
        protected int _HeaderLength;
        protected int _TextureLength;

        // Methods
        public SoulReaverMonolithicTextureFile(string path)
            : base(path)
        {
        }

        protected override int _GetTextureCount()
        {
            if (base._FileInfo.Length < (this._HeaderLength + this._TextureLength))
            {
                throw new TextureFileException(string.Concat(new object[] { "The file '", base._FilePath, "' does not contain enough data for it to be a texture file for the ", base._FileTypeName, " version of Soul Reaver (", this._HeaderLength, " byte header + ", this._TextureLength, " bytes per texture)." }));
            }
            long num = (base._FileInfo.Length - this._HeaderLength) / ((long)this._TextureLength);
            float num2 = (base._FileInfo.Length - this._HeaderLength) / ((float)this._TextureLength);
            if (num != num2)
            {
                throw new TextureFileException(string.Concat(new object[] { "The file '", base._FilePath, "' does not appear to be a valid texture file for the ", base._FileTypeName, " version of Soul Reaver. File lengths for this type should be ", this._HeaderLength, " bytes plus a whole number multiple of ", this._TextureLength, " bytes." }));
            }
            return (int)num;
        }

        public override void ExportFile(int index, string outPath)
        {
            base.GetTextureAsBitmap(index).Save(outPath, ImageFormat.Png);
        }
    }
}
