using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using Microsoft.DirectX.Direct3D;

namespace BenLincoln.TheLostWorlds.CDTextures
{
    public class SoulReaverPlaystationTextureFile : TextureFile
    {
        // Fields
        protected byte[][] _TextureData;
        protected Bitmap[] _Textures;
        protected int _TotalHeight;
        protected int _TotalWidth;

        // Methods
        public SoulReaverPlaystationTextureFile(string path)
            : base(path)
        {
            base._FileType = TextureFileType.SoulReaverPlaystation;
            base._FileTypeName = "Soul Reaver (Playstation) Indexed Texture File";
            Point imageSize = this.GetImageSize();
            this._TotalWidth = imageSize.X;
            this._TotalHeight = imageSize.Y;
            base._TextureCount = this._GetTextureCount();
            this._Textures = new Bitmap[base._TextureCount];
            this._TextureData = this.ReadTextureData(this.LoadTextureData());
        }

        protected override Texture _GetTexture(Device device, int index)
        {
            return Texture.FromBitmap(device, this._GetTextureAsBitmap(index), Usage.None, Pool.Managed);
        }

        protected override Bitmap _GetTextureAsBitmap(int index)
        {
            return this._Textures[index];
        }

        protected override int _GetTextureCount()
        {
            if (base._FileInfo.Length < 0x24L)
            {
                throw new TextureFileException("The file '" + base._FilePath + "' does not contain enough data for it to be a texture file for the Playstation version of Soul Reaver.");
            }
            long num = (base._FileInfo.Length - 0x24L) / 0x8000L;
            float num2 = (base._FileInfo.Length - 36f) / 32768f;
            if (num != num2)
            {
                num += 1L;
            }
            return (int)num;
        }

        public void BuildTexturesFromPolygonData(SoulReaverPlaystationPolygonTextureData[] texData, bool drawGreyScaleFirst, bool quantizeBounds)
        {
            Hashtable hashtable = new Hashtable();

            //#region AMF 24/03/08
            //bool[][][] unusedPixels = new bool[base._TextureCount][][];
            //for (int i = 0; i < base._TextureCount; i++)
            //{
            //    unusedPixels[i] = new bool[0x100][];
            //    for (int j = 0; j < 0x100; j++)
            //    {
            //        unusedPixels[i][j] = new bool[0x100];
            //    }
            //}
            //#endregion

            if (drawGreyScaleFirst)
            {
                for (int i = 0; i < base._TextureCount; i++)
                {
                    this._Textures[i] = this.GetTextureAsBitmap(i, this.GetGreyscalePalette());
                }
                return;
            }
            else
            {
                Color color = Color.FromArgb(0xFF, 0x80, 0x80, 0x80);
                Bitmap blankBitmap = new Bitmap(0x100, 0x100);
                for (int k = 0; k < 0x100; k++)
                {
                    for (int m = 0; m < 0x100; m++)
                    {
                        blankBitmap.SetPixel(m, k, color);
                    }
                }
                for (int j = 0; j < base._TextureCount; j++)
                {
                    this._Textures[j] = (Bitmap)blankBitmap.Clone();
                }
            }

            foreach (SoulReaverPlaystationPolygonTextureData data in texData)
            {
                string key = data.paletteColumn.ToString() + "-" + data.paletteRow.ToString();
                if (hashtable.Contains(key))
                {
                    int num5 = (int)hashtable[key];
                    num5++;
                    hashtable[key] = num5;
                }
                else
                {
                    int num6 = 1;
                    hashtable.Add(key, num6);
                }
                int uMin = 0xFF;
                int uMax = 0;
                int vMin = 0xFF;
                int vMax = 0;
                Color[] palette = this.GetPalette(data.paletteColumn, data.paletteRow);
                foreach (int u in data.u)
                {
                    uMin = Math.Min(uMin, u);
                    uMax = Math.Max(uMax, u);
                }
                foreach (int v in data.v)
                {
                    vMin = Math.Min(vMin, v);
                    vMax = Math.Max(vMax, v);
                }

                int width = 0x10;
                // width = uMin - uMax;
                // width = Math.Abs(uMin - uMax);
                //for (int b = 0; b < 8; b++)
                //{
                //    if ((1 << b) >= width)
                //    {
                //        width = 1 << b;
                //        break;
                //    }
                //}

                int height = 0x10;
                // height = vMin - vMax;
                // height = Math.Abs(vMin - vMax);
                //for (int b = 0; b < 8; b++)
                //{
                //    if ((1 << b) >= height)
                //    {
                //        height = 1 << b;
                //        break;
                //    }
                //}

                if (quantizeBounds)
                {
                    while ((uMin % width) > 0)
                    {
                        uMin--;
                    }
                    while ((uMax % width) > 0)
                    {
                        uMax++;
                    }
                    while ((vMin % height) > 0)
                    {
                        vMin--;
                    }
                    while ((vMax % height) > 0)
                    {
                        vMax++;
                    }
                    if (uMin < 0)
                    {
                        uMin = 0;
                    }
                    if (uMax > 0x100)
                    {
                        uMax = 0x100;
                    }
                    if (vMin < 0)
                    {
                        vMin = 0;
                    }
                    if (vMax > 0x100)
                    {
                        vMax = 0x100;
                    }
                }

                for (int n = vMin; n < vMax; n++)
                {
                    for (int num15 = uMin; num15 < uMax; num15 += 2)
                    {
                        int num16 = (n * 0x80) + (num15 / 2);
                        byte num17 = this._TextureData[data.textureID][num16];
                        int num18 = num17 & 15;
                        int num19 = (num17 & 240) >> 4;
                        Color color2 = palette[num18];
                        Color color3 = palette[num19];
                        this._Textures[data.textureID].SetPixel(num15, n, color2);
                        this._Textures[data.textureID].SetPixel(num15 + 1, n, color3);
                        //#region AMF 24/03/08
                        //unusedPixels[data.textureID][num15][n] = true;
                        //unusedPixels[data.textureID][num15 + 1][n] = true;
                        //#endregion
                    }
                }
            }

            //#region AMF 24/03/08
            //for (int t = 0; t < base._TextureCount; t++)
            //{
            //    for (int i = 0; i < 0x100; i++)
            //    {
            //        for (int j = 0; j < 0x100; j++)
            //        {
            //            if (unusedPixels[t][i][j] == false)
            //            {
            //                int _i = Math.Max(0, i);
            //                int _j = Math.Max(0, j);
            //                Color col00 = this._Textures[t].GetPixel(_i, _j);
            //                Color col01 = this._Textures[t].GetPixel(_i, _j);
            //                Color col10 = this._Textures[t].GetPixel(_i, _j);
            //                int A = (col00.A + col01.A + col10.A) / 3;
            //                int R = (col00.R + col01.R + col10.R) / 3;
            //                int G = (col00.G + col01.G + col10.G) / 3;
            //                int B = (col00.B + col01.B + col10.B) / 3;
            //                Color col11 = Color.FromArgb(A, R, B, G);
            //                this._Textures[t].SetPixel(i, j, col11);
            //            }
            //        }
            //    }
            //}
            //#endregion

            //int index = 0;
            //if (!drawGreyScaleFirst)
            //{
            //    string str2 = "";
            //    int num21 = 0;
            //    foreach (string str3 in hashtable.Keys)
            //    {
            //        int num22 = (int)hashtable[str3];
            //        if (num22 > num21)
            //        {
            //            str2 = str3;
            //        }
            //    }
            //    string[] strArray = str2.Split(new char[] { '-' });
            //    int paletteColumn = int.Parse(strArray[0]);
            //    int paletteRow = int.Parse(strArray[1]);
            //    Color[] colorArray2 = this.GetPalette(paletteColumn, paletteRow);
            //    for (index = 0; index <= this._Textures.GetUpperBound(0); index++)
            //    {
            //        for (int num25 = 0; num25 < 0x100; num25++)
            //        {
            //            for (int num26 = 0; num26 < 0x100; num26 += 2)
            //            {
            //                bool flag = this._Textures[index].GetPixel(num26, num25).A == 1;
            //                bool flag2 = this._Textures[index].GetPixel(num26 + 1, num25).A == 1;
            //                if (flag || flag2)
            //                {
            //                    int num27 = (num25 * 0x80) + (num26 / 2);
            //                    byte num28 = this._TextureData[index][num27];
            //                    int num29 = num28 & 15;
            //                    int num30 = (num28 & 240) >> 4;
            //                    Color color4 = colorArray2[num29];
            //                    Color color5 = colorArray2[num30];
            //                    if (flag)
            //                    {
            //                        this._Textures[index].SetPixel(num26, num25, color4);
            //                    }
            //                    if (flag2)
            //                    {
            //                        this._Textures[index].SetPixel(num26 + 1, num25, color5);
            //                    }
            //                }
            //            }
            //        }
            //    }
            //}
        }

        public override void ExportFile(int index, string outPath)
        {
            Bitmap textureAsBitmap = new Bitmap(1, 1);
            if (this._Textures[index] == null)
            {
                textureAsBitmap = this.GetTextureAsBitmap(index, this.GetGreyscalePalette());
            }
            else
            {
                textureAsBitmap = this._Textures[index];
            }
            textureAsBitmap.Save(outPath, ImageFormat.Png);
        }

        public override MemoryStream GetDataAsStream(int index)
        {
            throw new Exception("The method or operation is not implemented.");
        }

        protected Color[] GetGreyscalePalette()
        {
            Color[] colorArray = new Color[0x10];
            for (int i = 0; i < 0x10; i++)
            {
                int red = 0x10 * i;
                colorArray[i] = Color.FromArgb(red, red, red);
            }
            return colorArray;
        }

        protected Point GetImageSize()
        {
            Point point;
            try
            {
                FileStream input = new FileStream(base._FilePath, FileMode.Open, FileAccess.Read);
                BinaryReader reader = new BinaryReader(input);
                input.Seek(0x1cL, SeekOrigin.Begin);
                int x = (int)(reader.ReadUInt32() * 4);
                int y = (int)reader.ReadUInt32();
                reader.Close();
                input.Close();
                point = new Point(x, y);
            }
            catch (Exception exception)
            {
                throw new TextureFileException("Error getting image size.", exception);
            }
            return point;
        }

        protected Color[] GetPalette(int paletteColumn, int paletteRow)
        {
            Color[] colorArray = new Color[0x10];
            int index = paletteColumn / 4;
            int num2 = paletteColumn % 4;
            int num3 = (num2 * 0x20) + (paletteRow * 0x80);
            for (int i = 0; i < 0x10; i++)
            {
                int num5 = num3 + (i * 2);
                ushort num6 = (ushort)(this._TextureData[index][num5] | (this._TextureData[index][num5 + 1] << 8));
                ushort alpha = 0xff;
                ushort red = num6;
                ushort green = num6;
                ushort blue = num6;
                alpha = (ushort)(num6 >> 15);
                if (alpha != 0)
                {
                    alpha = 0xff;
                }
                blue = (ushort)((((ushort)(num6 << 1)) >> 11) << 3);
                green = (ushort)((((ushort)(num6 << 6)) >> 11) << 3);
                red = (ushort)((((ushort)(num6 << 11)) >> 11) << 3);
                colorArray[i] = Color.FromArgb(alpha, red, green, blue);
            }
            return colorArray;
        }

        protected Bitmap GetTextureAsBitmap(int index, Color[] palette)
        {
            Bitmap bitmap = new Bitmap(0x100, 0x100);
            int num = 0;
            for (int i = 0; i < 0x100; i++)
            {
                for (int j = 0; j < 0x100; j += 2)
                {
                    byte num4 = this._TextureData[index][num];
                    int num5 = num4 & 15;
                    int num6 = (num4 & 240) >> 4;
                    bitmap.SetPixel(j, i, palette[num5]);
                    bitmap.SetPixel(j + 1, i, palette[num6]);
                    num++;
                }
            }
            return bitmap;
        }

        protected byte[] LoadTextureData()
        {
            byte[] buffer;
            try
            {
                FileStream input = new FileStream(base._FilePath, FileMode.Open, FileAccess.Read);
                BinaryReader reader = new BinaryReader(input);
                //buffer = new byte[(this._TotalWidth / 2) * this._TotalHeight];
                input.Seek(0x24L, SeekOrigin.Begin);
                //for (int i = 0; i < buffer.Length; i++)
                //{
                //    buffer[i] = reader.ReadByte();
                //}
                buffer = reader.ReadBytes((this._TotalWidth / 2) * this._TotalHeight);
                reader.Close();
                input.Close();
            }
            catch (Exception exception)
            {
                throw new TextureFileException("Error reading texture.", exception);
            }
            return buffer;
        }

        protected byte[][] ReadTextureData(byte[] inData)
        {
            int num = inData.Length / 0x8000;
            int num2 = inData.Length % 0x8000;
            if (num2 > 0)
            {
                num++;
            }
            byte[][] bufferArray = new byte[num][];
            for (int i = 0; i < num; i++)
            {
                bufferArray[i] = new byte[0x8000];
            }
            int index = 0;
            for (int j = 0; j < this._TotalHeight; j++)
            {
                for (int k = 0; k < num; k++)
                {
                    int num7 = 0x80 * k;
                    int num8 = num7 + 0x80;
                    int num9 = (this._TotalWidth / 2) - num7;
                    int num10 = 0;
                    if (num9 < 0x80)
                    {
                        num10 = 0x80 - num9;
                        num8 = num7 + num9;
                    }
                    for (int m = num7; m < num8; m++)
                    {
                        bufferArray[k][(j * 0x80) + (m - num7)] = inData[index];
                        index++;
                    }
                    for (int n = 0; n < num10; n++)
                    {
                        bufferArray[k][((j * 0x80) + (num8 - num7)) + n] = 0;
                    }
                }
            }
            return bufferArray;
        }

        // Nested Types
        [StructLayout(LayoutKind.Sequential)]
        public struct SoulReaverPlaystationPolygonTextureData
        {
            public int[] u;
            public int[] v;
            public int textureID;
            public int paletteColumn;
            public int paletteRow;
        }
    }
}
