using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace ModelEx3
{
    class ModelExCamera
    {
        public enum CameraMode
        {
            Manual = 0,
            Fly = 1,
            Orbit = 2,
            Map = 3,
        };

        Device                  m_xDevice;
        Control                 m_xRenderWindow;
        Vector3                 m_xPosition;
        Vector3                 m_xTarget;
        Vector3                 m_xUp;
        Quaternion              m_xRotation;
        ModelExMouseController  m_xArcBall;
        CameraMode              m_eMode;

        float                   m_fDistance;
        float                   m_fFieldOfView;
        float                   m_fAspectRatio;
        float                   m_fNearClipPlane;
        float                   m_fFarClipPlane;

        public Vector3          Position
        {
            get { return m_xPosition; }
            set
            {
                m_xPosition = value;
                Matrix xRotation = Matrix.RotationQuaternion(m_xRotation);

                Vector3 xBaseForward = new Vector3(0.0f, 0.0f, m_fDistance);
                xBaseForward.TransformCoordinate(xRotation);
                m_xTarget = m_xPosition + xBaseForward;

                Vector3 xBaseUp = new Vector3(0.0f, 1.0f, 0.0f);
                xBaseUp.TransformCoordinate(xRotation);
                m_xUp = xBaseUp;

                //Vector3 xDirection = m_xTarget - m_xPosition;
                //m_fDistance = Math.Abs(xDirection.Length());
                //Matrix xRotation = Matrix.LookAtLH(
                //    Vector3.Empty,
                //    xDirection,
                //    new Vector3(0.0f, 1.0f, 0.0f)
                //);
                //m_xRotation = Quaternion.RotationMatrix(xRotation);
                //m_xArcBall.Rotation = m_xRotation;
                //UpdateVectors();
            }
        }
        public Vector3          Target
        {
            get { return m_xTarget; }
            set
            {
                m_xTarget = value;
                Matrix xRotation = Matrix.RotationQuaternion(m_xRotation);

                Vector3 xBaseForward = new Vector3(0.0f, 0.0f, m_fDistance);
                xBaseForward.TransformCoordinate(xRotation);
                m_xPosition = m_xTarget - xBaseForward;

                Vector3 xBaseUp = new Vector3(0.0f, 1.0f, 0.0f);
                xBaseUp.TransformCoordinate(xRotation);
                m_xUp = xBaseUp;

                //Vector3 xDirection = m_xTarget - m_xPosition;
                //m_fDistance = Math.Abs(xDirection.Length());
                //Matrix xRotation = Matrix.LookAtLH(
                //    Vector3.Empty,
                //    xDirection,
                //    new Vector3(0.0f, 1.0f, 0.0f)
                //);
                //m_xRotation = Quaternion.RotationMatrix(xRotation);
                //m_xArcBall.Rotation = m_xRotation;
                //UpdateVectors();
            }
        }
        public float            Distance
        {
            get { return m_fDistance; }
            set
            {
                m_fDistance = Math.Max(m_fNearClipPlane, value);
                UpdateVectors();
            }
        }
        public Quaternion       Rotation
        {
            get { return m_xRotation; }
            set
            {
                m_xRotation = value;
                UpdateVectors();
            }
        }
        public ModelExArcBall   ArcBall
        {
            get { return m_xArcBall; }
        }
        public CameraMode       Mode
        {
            get
            {
                return m_eMode;
            }
            set
            {
                if (m_eMode != value)
                {
                    switch (value)
                    {
                        case CameraMode.Map:
                            m_xUp = new Vector3(0.0f, 0.0f, 1.0f);
                            m_xPosition = new Vector3(0.0f, 2.5f, 0.0f);
                            m_xTarget = new Vector3(0.0f, 0.0f, 0.0f);
                            m_fDistance = 2.5f;
                            Matrix xRotation = Matrix.LookAtLH(
                                m_xPosition,
                                m_xTarget,
                                m_xUp
                            );
                            m_xRotation = Quaternion.RotationMatrix(xRotation);
                            m_xRotation.Invert();
                            m_xArcBall.Rotation = m_xRotation;
                            break;
                        case CameraMode.Fly:
                            m_xArcBall.Mode = ModelExMouseController.ControlMode.XY;
                            RemoveRoll();
                            break;
                        default:
                            m_xArcBall.Mode = ModelExMouseController.ControlMode.XYZ;
                            break;
                    }
                    m_xArcBall.Rotation = m_xRotation;
                }
                m_eMode = value;
            }
        }

        public ModelExCamera(ref Device xDevice, Vector3 xTarget, float fDistance)
        {
            m_xDevice               = xDevice;
            m_eMode                 = CameraMode.Orbit;
            m_xTarget               = xTarget;
            m_xUp                   = new Vector3(0.0f, 1.0f, 0.0f);
            m_fDistance             = fDistance;
            m_xRotation             = Quaternion.Identity;

            Matrix xRotation        = Matrix.RotationQuaternion(m_xRotation);
            Vector3 xTargetVector   = new Vector3(0.0f, 0.0f, m_fDistance);
            xTargetVector.TransformCoordinate(xRotation);
            m_xPosition             = m_xTarget - xTargetVector;

            m_xRenderWindow         = xDevice.PresentationParameters.DeviceWindow;
            m_fFieldOfView          = (float)Math.PI / 4.0f;
            m_fAspectRatio          = (float)m_xRenderWindow.Width / (float)m_xRenderWindow.Height;
            m_fNearClipPlane        = 0.1f;
            m_fFarClipPlane         = 10000.0f;

            m_xArcBall              = new ModelExMouseController(ref m_xRenderWindow, true);
            m_xArcBall.Move        += new ModelExArcBall.MoveHandler(ModelExCamera_Drag);
        }

        public void MoveForward(float fDistance)
        {
            Vector3 xBaseForward = new Vector3(0.0f, 0.0f, 1.0f);
            xBaseForward.TransformCoordinate(Matrix.RotationQuaternion(m_xRotation));
            xBaseForward *= fDistance;
            m_xPosition += xBaseForward;
            m_xTarget += xBaseForward;
        }

        public void MoveUpward(float fDistance)
        {
            Vector3 xBaseUp = new Vector3(0.0f, 1.0f, 0.0f);
            xBaseUp.TransformCoordinate(Matrix.RotationQuaternion(m_xRotation));
            xBaseUp *= fDistance;
            m_xPosition += xBaseUp;
            m_xTarget += xBaseUp;
        }

        public void MoveRight(float fDistance)
        {
            Vector3 xBaseRight = new Vector3(1.0f, 0.0f, 0.0f);
            xBaseRight.TransformCoordinate(Matrix.RotationQuaternion(m_xRotation));
            xBaseRight *= fDistance;
            m_xPosition += xBaseRight;
            m_xTarget += xBaseRight;
        }

        public void RotateX(float fAngle)
        {
            Vector3 xBaseUp = new Vector3(0.0f, 1.0f, 0.0f);
            Quaternion xRotationAxis = Quaternion.RotationAxis(xBaseUp, fAngle);
            if (m_eMode == CameraMode.Orbit)
            {
                xRotationAxis.Invert();
                m_xRotation = xRotationAxis * m_xRotation;
            }
            else
            {
                m_xRotation = m_xRotation * xRotationAxis;
            }

            m_xRotation.Normalize();
            m_xArcBall.Rotation = m_xRotation;
            UpdateVectors();
        }

        public void RotateY(float fAngle)
        {
            Vector3 xBaseRight = new Vector3(1.0f, 0.0f, 0.0f);
            Quaternion xRotationAxis = Quaternion.RotationAxis(xBaseRight, fAngle);
            if (m_eMode == CameraMode.Orbit)
            {
                xRotationAxis.Invert();
            }
            m_xRotation = xRotationAxis * m_xRotation;
            m_xRotation.Normalize();
            m_xArcBall.Rotation = m_xRotation;
            UpdateVectors();
        }

        public void RotateZ(float fAngle)
        {
            Vector3 xBaseForward = new Vector3(0.0f, 0.0f, 1.0f);
            Quaternion xRotationAxis = Quaternion.RotationAxis(xBaseForward, fAngle);
            if (m_eMode == CameraMode.Orbit)
            {
                xRotationAxis.Invert();
            }
            m_xRotation = xRotationAxis * m_xRotation;
            m_xRotation.Normalize();
            m_xArcBall.Rotation = m_xRotation;
            UpdateVectors();
        }

        public void RemoveRoll()
        {
            m_xUp = new Vector3(0.0f, 1.0f, 0.0f);
            Matrix xRotation = Matrix.LookAtLH(
                m_xPosition,
                m_xTarget,
                m_xUp
            );
            m_xRotation = Quaternion.RotationMatrix(xRotation);
            m_xRotation.Invert();
            m_xArcBall.Rotation = m_xRotation;
        }

        protected void UpdateVectors()
        {
            Matrix xRotation = Matrix.RotationQuaternion(m_xRotation);

            Vector3 xBaseForward = new Vector3(0.0f, 0.0f, m_fDistance);
            xBaseForward.TransformCoordinate(xRotation);

            if (Mode == CameraMode.Fly || Mode == CameraMode.Map)
            {
                m_xTarget = m_xPosition + xBaseForward;

            }
            else if (Mode == CameraMode.Orbit)
            {
                m_xPosition = m_xTarget - xBaseForward;
            }

            Vector3 xBaseUp = new Vector3(0.0f, 1.0f, 0.0f);
            xBaseUp.TransformCoordinate(xRotation);
            m_xUp = xBaseUp;
        }

        protected void ModelExCamera_Drag()
        {
            m_xRotation = m_xArcBall.Rotation;
            UpdateVectors();
        }

        public void Update()
        {
            m_fAspectRatio = (float)m_xRenderWindow.Width / (float)m_xRenderWindow.Height;

            if (m_eMode == CameraMode.Map)
            {
                Vector3 xDirection = Target - Position;
                float fDistance = Math.Abs(xDirection.Length());
                float fHeight = 2.0f * fDistance * (float)Math.Tan(m_fFieldOfView / 2.0f);
                float fWidth = fHeight * m_fAspectRatio;

                m_xDevice.Transform.Projection = Matrix.OrthoLH(
                    fWidth,
                    fHeight,
                    m_fNearClipPlane,
                    m_fFarClipPlane
                );
            }
            else
            {
                m_xDevice.Transform.Projection = Matrix.PerspectiveFovLH(
                    m_fFieldOfView,
                    m_fAspectRatio,
                    m_fNearClipPlane,
                    m_fFarClipPlane
                );
            }

            m_xDevice.Transform.View = Matrix.LookAtLH(
                m_xPosition,
                m_xTarget,
                m_xUp
            );
        }
    }
}