using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Text;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using BenLincoln.TheLostWorlds.CDTextures;
using CDTextureFile = BenLincoln.TheLostWorlds.CDTextures.TextureFile;
using SR1PCTextureFile = BenLincoln.TheLostWorlds.CDTextures.SoulReaverPCTextureFile;
using SR1PSTextureFile = BenLincoln.TheLostWorlds.CDTextures.SoulReaverPlaystationTextureFile;
using SR1DCTextureFile = BenLincoln.TheLostWorlds.CDTextures.SoulReaverDreamcastTextureFile;
using VertexPNT = Microsoft.DirectX.Direct3D.CustomVertex.PositionNormalTextured;
using VertexPCT = Microsoft.DirectX.Direct3D.CustomVertex.PositionColoredTextured;

namespace ModelEx3
{
    class ModelExSRObjectInstance : ModelExObjectInstance
    {
        SR1File m_xSR1File;

        public override String TextureArchives
        {
            get
            {
                if (m_xSR1File == null)
                {
                    return null;
                }

                switch (m_xSR1File.m_ePlatform)
                {
                    case Platform.PC:
                        return
                            "Soul Reaver Monolithic Files (*.big)|*.big|" +
                            "All Files (*.*)|*.*";
                    case Platform.PSX:
                        return
                            "Soul Reaver CRM Files (*.crm)|*.crm|" +
                            //"Soul Reaver Archive Files (*.dat)|*.dat|" +
                            "All Files (*.*)|*.*";
                    case Platform.Dreamcast:
                        return
                            "Soul Reaver Archive Files (*.vq)|*.vq|" +
                            "All Files (*.*)|*.*";
                    default:
                        return null;
                }
            }
        }

        public override String TextureArchiveExt
        {
            get
            {
                switch (m_xSR1File.m_ePlatform)
                {
                    case Platform.PC: return "big";
                    case Platform.PSX: return "crm";
                    case Platform.Dreamcast: return "vq";
                    default: return null;
                }
            }
        }

        public ModelExSRObjectInstance(ModelExSystem xSystem, String strFilePath)
            : base(xSystem, strFilePath)
        {
        }

        ~ModelExSRObjectInstance()
        {
            m_xSR1File = null;
        }

        public override void Load(string strFilePath)
        {
            try
            {
                base.Load(strFilePath);
            }
            catch
            {
                m_xModels.List.Clear();
                m_xEffects.List.Clear();
                m_xTextures.List.Clear();
                m_xContainers.Clear();

                m_xSR1File = new SR1File(strFilePath);

                for (int iModel = 0; iModel < m_xSR1File.m_usModelCount; iModel++)
                {
                    Char[] acModelName = m_xSR1File.m_strModelName.ToCharArray();
                    acModelName[0] = Char.ToUpper(acModelName[0]);
                    String strModelName = new String(acModelName);

                    ModelExModelInstance xModel = new ModelExModelInstance(m_xSystem, strModelName + "_" + iModel.ToString() + ".x");

                    ModelExFrame xFrame = new ModelExFrame();
                    xFrame.Name = "Root";
                    xFrame.TransformationMatrix = Matrix.Identity;

                    // Create the container
                    xFrame.ModelExContainer = new ModelExContainer();
                    ModelExContainer xContainer = xFrame.ModelExContainer;
                    xContainer.Name = strModelName + "_" + iModel.ToString();

                    // Set the vertex format
                    VertexFormats eVertexFormat = (m_xSR1File.m_eFileType == FileType.Object) ?
                        VertexPNT.Format :
                        VertexPCT.Format;

                    // Create the mesh
                    xContainer.Mesh = new Mesh(
                        (int)m_xSR1File.m_axModels[iModel].PolygonCount,
                        (int)m_xSR1File.m_axModels[iModel].IndexCount,
                        MeshFlags.Managed,
                        eVertexFormat,
                        m_xSystem.Device
                    );

                    // Insert the vertices
                    if (eVertexFormat == VertexPNT.Format)
                    {
                        VertexPNT[] xVerticesPNT = new VertexPNT[m_xSR1File.m_axModels[iModel].IndexCount];
                        for (UInt16 v = 0; v < m_xSR1File.m_axModels[iModel].IndexCount; v++)
                        {
                            Vector3 xNormal = new Vector3();
                            xNormal.X = (float)m_xSR1File.m_axModels[iModel].Vertices[v].normal.x;
                            xNormal.Y = (float)m_xSR1File.m_axModels[iModel].Vertices[v].normal.z;
                            xNormal.Z = (float)m_xSR1File.m_axModels[iModel].Vertices[v].normal.y;
                            xNormal.Normalize();

                            xVerticesPNT[v] = new VertexPNT(
                                0.001f * (float)m_xSR1File.m_axModels[iModel].Vertices[v].worldPos.x, // 100.0f
                                0.001f * (float)m_xSR1File.m_axModels[iModel].Vertices[v].worldPos.z, // 100.0f
                                0.001f * (float)m_xSR1File.m_axModels[iModel].Vertices[v].worldPos.y, // 100.0f
                                xNormal.X, // (float)m_xSR1File.m_axModels[iModelID].Vertices[v].normal.x,
                                xNormal.Y, // (float)m_xSR1File.m_axModels[iModelID].Vertices[v].normal.z,
                                xNormal.Z, // (float)m_xSR1File.m_axModels[iModelID].Vertices[v].normal.y,
                                m_xSR1File.m_axModels[iModel].Vertices[v].fU,
                                m_xSR1File.m_axModels[iModel].Vertices[v].fV
                            );
                        }
                        xContainer.Mesh.VertexBuffer.SetData(xVerticesPNT, 0, LockFlags.None);
                    }
                    else
                    {
                        VertexPCT[] xVerticesPCT = new VertexPCT[m_xSR1File.m_axModels[iModel].IndexCount];
                        for (UInt16 v = 0; v < m_xSR1File.m_axModels[iModel].IndexCount; v++)
                        {
                            xVerticesPCT[v] = new VertexPCT(
                                0.001f * (float)m_xSR1File.m_axModels[iModel].Vertices[v].worldPos.x,
                                0.001f * (float)m_xSR1File.m_axModels[iModel].Vertices[v].worldPos.z,
                                0.001f * (float)m_xSR1File.m_axModels[iModel].Vertices[v].worldPos.y,
                                (int)m_xSR1File.m_axModels[iModel].Vertices[v].colour,
                                m_xSR1File.m_axModels[iModel].Vertices[v].fU,
                                m_xSR1File.m_axModels[iModel].Vertices[v].fV
                            );
                        }
                        xContainer.Mesh.VertexBuffer.SetData(xVerticesPCT, 0, LockFlags.None);
                    }

                    // Insert the indices
                    UInt16[] iIndices = new UInt16[m_xSR1File.m_axModels[iModel].IndexCount];
                    for (UInt16 i = 0; i < m_xSR1File.m_axModels[iModel].IndexCount; i++)
                    {
                        if (m_xSR1File.m_axModels[iModel].Indices[i] % 3 == 0)
                        {
                            iIndices[i] = m_xSR1File.m_axModels[iModel].Indices[i];
                        }
                        if (m_xSR1File.m_axModels[iModel].Indices[i] % 3 == 1)
                        {
                            iIndices[i] = m_xSR1File.m_axModels[iModel].Indices[i + 1];
                        }
                        if (m_xSR1File.m_axModels[iModel].Indices[i] % 3 == 2)
                        {
                            iIndices[i] = m_xSR1File.m_axModels[iModel].Indices[i - 1];
                        }
                    }
                    xContainer.Mesh.IndexBuffer.SetData(iIndices, 0, LockFlags.None);

                    // Get the attributes
                    Int32[] axAttributeArray = xContainer.Mesh.LockAttributeBufferArray(LockFlags.Discard);
                    for (UInt16 p = 0; p < m_xSR1File.m_axModels[iModel].PolygonCount; p++)
                    {
                        axAttributeArray[p] = m_xSR1File.m_axModels[iModel].Polygons[p].material.ID;
                    }
                    xContainer.Mesh.UnlockAttributeBuffer(axAttributeArray);

                    // Weld any duplicate vertices
                    xContainer.Mesh.WeldVertices(
                        WeldEpsilonsFlags.WeldPartialMatches,
                        new WeldEpsilons(),
                        null
                    );

                    // Generate the adjacency
                    int[] aiAdjacency = new int[xContainer.Mesh.NumberFaces * 3];
                    xContainer.Mesh.GenerateAdjacency(0.0f, aiAdjacency);
                    xContainer.SetAdjacency(aiAdjacency);

                    // Compute the normals
                    if (m_xSR1File.m_eFileType == FileType.Area)
                    {
                        xContainer.Mesh = xContainer.Mesh.Clone(
                            MeshFlags.Managed,
                            xContainer.Mesh.VertexFormat | VertexFormats.Normal,
                            m_xSystem.Device
                        );
                        xContainer.Mesh.ComputeNormals(aiAdjacency);
                    }

                    // Compute the bounding sphere
                    GraphicsStream xVertexData = xContainer.Mesh.VertexBuffer.Lock(0, 0, LockFlags.None);
                    Vector3 xCentre = Vector3.Empty;
                    float fRadius = Geometry.ComputeBoundingSphere(
                        xVertexData,
                        xContainer.Mesh.NumberVertices,
                        xContainer.Mesh.VertexFormat,
                        out xCentre
                    );
                    xContainer.Mesh.VertexBuffer.Unlock();
                    xContainer.Centre = xCentre;
                    xContainer.Radius = fRadius;

                    float fScale = 1.0f / fRadius;
                    xContainer.Scale = new Vector3(fScale, fScale, fScale);

                    // Create the materials
                    ExtendedMaterial[] axExtendedMaterials = axExtendedMaterials = new ExtendedMaterial[m_xSR1File.m_axModels[iModel].MaterialCount];
                    int[] aiTextures = new int[m_xSR1File.m_axModels[iModel].MaterialCount];
                    for (int iMaterial = 0; iMaterial < m_xSR1File.m_axModels[iModel].MaterialCount; iMaterial++)
                    {
                        axExtendedMaterials[iMaterial] = CreateMaterial(iModel, iMaterial);
                        String strTextureName = axExtendedMaterials[iMaterial].TextureFilename;

                        aiTextures[iMaterial] = -1;
                        for (int t = 0; t < m_xTextures.Count; t++)
                        {
                            if (m_xTextures[t].Name == strTextureName)
                            {
                                aiTextures[iMaterial] = t;
                                break;
                            }
                        }

                        if (aiTextures[iMaterial] == -1)
                        {
                            ModelExTextureInstance xTexture = new ModelExTextureInstance(m_xSystem, strTextureName);
                            xTexture.Load("");
                            m_xTextures.List.Add(xTexture);
                            aiTextures[iMaterial] = m_xTextures.Count - 1;
                        }
                    }
                    xContainer.SetMaterials(axExtendedMaterials);
                    xContainer.Textures = aiTextures;

                    // Create the effects
                    EffectInstance[] axEffectInstances = new EffectInstance[m_xSR1File.m_axModels[iModel].MaterialCount];
                    int[] aiEffects = new int[m_xSR1File.m_axModels[iModel].MaterialCount];
                    for (int iEffect = 0; iEffect < m_xSR1File.m_axModels[iModel].MaterialCount; iEffect++)
                    {
                        axEffectInstances[iEffect] = CreateEffectInstance(iModel, iEffect);
                        String strEffectName = axEffectInstances[iEffect].EffectFilename;

                        aiEffects[iEffect] = -1;
                        for (int e = 0; e < m_xEffects.Count; e++)
                        {
                            if (m_xEffects[e].Name == axEffectInstances[iEffect].EffectFilename)
                            {
                                aiEffects[iEffect] = e;
                                break;
                            }
                        }

                        if (aiEffects[iEffect] == -1)
                        {
                            ModelExEffectInstance xEffect = new ModelExEffectInstance(m_xSystem, axEffectInstances[iEffect]);
                            xEffect.Load(xEffect.Name);
                            m_xEffects.List.Add(xEffect);
                            aiEffects[iEffect] = m_xEffects.Count - 1;
                        }
                    }
                    xContainer.SetEffectInstances(axEffectInstances);
                    xContainer.SetEffectInstancesBackup(axEffectInstances);
                    xContainer.Effects = aiEffects;

                    m_xContainers.Add(xContainer);

                    ModelExModelFile xModelFile = ModelExModelFile.Create(m_xSystem, xModel.Name, xFrame);
                    xModel.Load(xModelFile);

                    m_xModels.List.Add(xModel);
                }

                // LoadTexturesFromPCMonolithicFile("C:/Documents and Settings/Andrew/My Documents/KainFiles/SR Executables/PC/textures.big");

                ModelIndex = 0;
            }
        }

        public override void Save(string strFilePath, ModelFormat eFormat)
        {
            base.Save(strFilePath, eFormat);
        }

        protected ExtendedMaterial CreateMaterial(int iModel, int iMaterial)
        {
            Material xMaterial = new Material();

            if (m_xSR1File.m_eFileType == FileType.Object &&
                m_xSR1File.m_axModels[iModel].Materials[iMaterial].textureUsed)
            {
                xMaterial.Diffuse = Color.White;
                xMaterial.Ambient = Color.FromArgb(unchecked((int)0xFF404040));
            }
            else
            {
                xMaterial.Diffuse = Color.FromArgb(
                    (Int32)(m_xSR1File.m_axModels[iModel].Materials[iMaterial].colour)
                );
            }

            String strTextureName = m_xSR1File.m_axModels[iModel].Materials[iMaterial].textureName;
            ExtendedMaterial xExtendedMaterial = new ExtendedMaterial();
            xExtendedMaterial.Material3D = xMaterial;
            xExtendedMaterial.TextureFilename = strTextureName;

            return xExtendedMaterial;
        }

        protected EffectInstance CreateEffectInstance(int iModel, int iMaterial)
        {
            String strTextureName   = m_xSR1File.m_axModels[iModel].Materials[iMaterial].textureName;
            int iColourOperaton     = (int)TextureOperation.Modulate;
            int iLighting           = 1;
            int iAmbientSource      = (int)ColorSource.Material;
            int iDiffuseSource      = (int)ColorSource.Color1;
            int iColourVertex       = 0;

            if (m_xSR1File.m_eFileType == FileType.Area)
            {
                iColourOperaton = (int)TextureOperation.Modulate4X;
                iLighting       = 0;
                iAmbientSource  = (int)ColorSource.Color1;
                iColourVertex   = 1;
            }

            EffectDefault[] axEffectDefaults = new EffectDefault[7];

            axEffectDefaults[0].ParameterName = "xTexture0";
            axEffectDefaults[0].Type = EffectDefaultType.String;
            axEffectDefaults[0].Data = strTextureName;

            axEffectDefaults[1].ParameterName = "xDiffuse";
            axEffectDefaults[1].Type = EffectDefaultType.Floats;
            float[] afDiffuse = { 1.0f, 1.0f, 1.0f, 1.0f };
            axEffectDefaults[1].Data = afDiffuse;

            axEffectDefaults[2].ParameterName = "xLighting";
            axEffectDefaults[2].Type = EffectDefaultType.Integer;
            axEffectDefaults[2].Data = iLighting;

            axEffectDefaults[3].ParameterName = "xColorVertex";
            axEffectDefaults[3].Type = EffectDefaultType.Integer;
            axEffectDefaults[3].Data = iColourVertex;

            axEffectDefaults[4].ParameterName = "xColorOperation";
            axEffectDefaults[4].Type = EffectDefaultType.Integer;
            axEffectDefaults[4].Data = iColourOperaton;

            axEffectDefaults[5].ParameterName = "xAmbientMaterialSource";
            axEffectDefaults[5].Type = EffectDefaultType.Integer;
            axEffectDefaults[5].Data = iAmbientSource;

            axEffectDefaults[6].ParameterName = "xDiffuseMaterialSource";
            axEffectDefaults[6].Type = EffectDefaultType.Integer;
            axEffectDefaults[6].Data = iDiffuseSource;

            String strFileName = "SoulReaver1Default.fx";
            EffectInstance xEffectInstance = new EffectInstance();
            xEffectInstance.SetDefaults(axEffectDefaults);
            xEffectInstance.EffectFilename = strFileName;

            return xEffectInstance;
        }

        public override void LoadTexturesFromFile(String strFileName)
        {
            CDTextureFile xCDTextureFile = null;
            switch (m_xSR1File.m_ePlatform)
            {
                case Platform.PC:
                    xCDTextureFile = LoadPCMonolithicFile(strFileName);
                    break;
                case Platform.PSX:
                    xCDTextureFile = LoadCRMFile(strFileName);
                    break;
                case Platform.Dreamcast:
                    xCDTextureFile = LoadDCMonolithicFile(strFileName);
                    break;
            }

            m_xTextures.List.Clear();
            for (int iModel = 0; iModel < m_xSR1File.m_usModelCount; iModel++)
            {
                int[] aiTextures = new int[m_xSR1File.m_axModels[iModel].MaterialCount];
                for (int iMaterial = 0; iMaterial < m_xSR1File.m_axModels[iModel].MaterialCount; iMaterial++)
                {
                    String strTextureName = m_xSR1File.m_axModels[iModel].Materials[iMaterial].textureName;
                    int iTextureID = m_xSR1File.m_axModels[iModel].Materials[iMaterial].textureID;

                    aiTextures[iMaterial] = -1;
                    for (int t = 0; t < m_xTextures.Count; t++)
                    {
                        if (m_xTextures[t].FileName == strTextureName)
                        {
                            aiTextures[iMaterial] = t;
                            break;
                        }
                    }

                    if (aiTextures[iMaterial] == -1)
                    {
                        ModelExTextureInstance xTexture = new ModelExTextureInstance(m_xSystem, strTextureName);
                        Texture xDXTexture = xCDTextureFile.GetTexture(m_xSystem.Device, iTextureID);
                        ModelExTextureFile xTextureFile = ModelExTextureFile.Create(m_xSystem, strTextureName, xDXTexture);
                        xTexture.Load(strTextureName);
                        m_xTextures.List.Add(xTexture);
                        aiTextures[iMaterial] = m_xTextures.Count - 1;
                    }
                }
                m_xContainers[iModel].Textures = aiTextures;
            }
        }

        protected CDTextureFile LoadPCMonolithicFile(String strFilePath)
        {
            return new SoulReaverPCTextureFile(strFilePath);
        }

        protected CDTextureFile LoadCRMFile(String strFilePath)
        {
            SR1PSTextureFile.SoulReaverPlaystationPolygonTextureData[] xPolygons =
                new SR1PSTextureFile.SoulReaverPlaystationPolygonTextureData[m_xSR1File.m_axModels[0].PolygonCount];

            int polyNum = 0;
            foreach (ExPolygon xPolygon in m_xSR1File.m_axModels[0].Polygons)
            {
                xPolygons[polyNum].paletteColumn = xPolygon.paletteColumn;
                xPolygons[polyNum].paletteRow = xPolygon.paletteRow;
                xPolygons[polyNum].u = new int[3];
                xPolygons[polyNum].v = new int[3];
                xPolygons[polyNum].u[0] = (int)xPolygon.v1.u;
                xPolygons[polyNum].u[1] = (int)xPolygon.v2.u;
                xPolygons[polyNum].u[2] = (int)xPolygon.v3.u;
                xPolygons[polyNum].v[0] = (int)xPolygon.v1.v;
                xPolygons[polyNum].v[1] = (int)xPolygon.v2.v;
                xPolygons[polyNum].v[2] = (int)xPolygon.v3.v;
                xPolygons[polyNum].textureID = xPolygon.material.textureID;
                polyNum++;
            }

            SR1PSTextureFile xSR1TextureFile = new SR1PSTextureFile(strFilePath);
            xSR1TextureFile.BuildTexturesFromPolygonData(xPolygons, false, true);
            return xSR1TextureFile;
        }

        protected CDTextureFile LoadDCMonolithicFile(String strFilePath)
        {
            return new SoulReaverDreamcastTextureFile(strFilePath);
        }
    }
}