using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;
using Imaging = System.Drawing.Imaging;

namespace ModelEx3
{
    class ModelExTextureFile : ModelExFile
    {
        protected ModelExSystem m_xSystem;
        protected String m_strFileName;
        protected Texture m_xTexture;

        public String FileName { get { return m_strFileName; } }
        public Texture Texture { get { return m_xTexture; } }

        protected ModelExTextureFile(ModelExSystem xSystem, String strFilePath)
        {
            m_xSystem = xSystem;
            m_strFileName = strFilePath;

            if (m_xSystem == null)
            {
                throw new Exception("Texture file must have a system");
            }

            if (m_strFileName == null)
            {
                throw new Exception("Texture file must have a folder");
            }

            if (m_strFileName == null)
            {
                throw new Exception("Texture file must have a name");
            }

            Load(strFilePath);
        }

        protected ModelExTextureFile(ModelExSystem xSystem, String strFilePath, Texture xTexture)
        {
            m_xSystem = xSystem;
            m_strFileName = strFilePath;

            if (m_xSystem == null)
            {
                throw new Exception("Texture file must have a system");
            }

            if (m_strFileName == null)
            {
                throw new Exception("Texture file must have a folder");
            }

            if (m_strFileName == null)
            {
                throw new Exception("Texture file must have a name");
            }

            Load(xTexture);
        }

        public static ModelExTextureFile Create(ModelExSystem xSystem, String strFilePath)
        {
            ModelExTextureFile xTextureFile = xSystem.FindTextureFile(strFilePath);

            if (xTextureFile == null)
            {
                xTextureFile = new ModelExTextureFile(xSystem, strFilePath);
                xSystem.AddTextureFile(xTextureFile);
            }

            return xTextureFile;
        }

        public static ModelExTextureFile Create(ModelExSystem xSystem, String strFilePath, Texture xTexture)
        {
            ModelExTextureFile xTextureFile = xSystem.FindTextureFile(strFilePath);

            if (xTextureFile == null)
            {
                xTextureFile = new ModelExTextureFile(xSystem, strFilePath, xTexture);
                xSystem.AddTextureFile(xTextureFile);
            }

            return xTextureFile;
        }

        public void Load(String strFilePath)
        {
            Bitmap xBitmap = new Bitmap(strFilePath);
            m_xTexture = new Texture(
                m_xSystem.Device,
                xBitmap,
                Usage.None,
                Pool.Managed
            );
        }

        public void Load(Texture xTexture)
        {
            m_xTexture = xTexture;
        }

        public void Save(String strFilePath, ImageFormat eFormat)
        {
            GraphicsStream xTextureStream = m_xTexture.LockRectangle(0, LockFlags.ReadOnly);
            SurfaceDescription xTextureDesc = m_xTexture.GetLevelDescription(0);
            Bitmap xTextureImage = new Bitmap(
                xTextureDesc.Width,
                xTextureDesc.Height,
                xTextureDesc.Width * 4,
                System.Drawing.Imaging.PixelFormat.Format32bppArgb,
                xTextureStream.InternalData
            );
            m_xTexture.UnlockRectangle(0);

            Imaging.ImageFormat xFormat = null;
            switch (eFormat)
            {
                case ImageFormat.Bmp:
                    xFormat = Imaging.ImageFormat.Bmp;
                    break;
                case ImageFormat.Jpeg:
                    xFormat = Imaging.ImageFormat.Jpeg;
                    break;
                case ImageFormat.Png:
                    xFormat = Imaging.ImageFormat.Png;
                    break;
            }

            xTextureImage.Save(strFilePath, xFormat);
            xTextureStream.Close();
        }
    }

    public abstract class ModelExTexture
    {
        public abstract String Name { get; }
        public abstract String FileName { get; }

        public abstract void Save(String strFilePath, ImageFormat eFormat);
    }

    class ModelExTextureList : ModelExList<ModelExTexture>
    {
        List<ModelExTextureInstance> m_xList;

        public List<ModelExTextureInstance> List { get { return m_xList; } }
        public int Count { get { return m_xList.Count; } }
        public ModelExTexture this[int t] { get { return m_xList[t]; } }

        public ModelExTextureList()
        {
            m_xList = new List<ModelExTextureInstance>();
        }

        public IEnumerator<ModelExTexture> GetEnumerator()
        {
            for (int t = 0; t < Count; t++)
            {
                yield return m_xList[t];
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
    }

    class ModelExTextureInstance : ModelExTexture
    {
        protected ModelExSystem m_xSystem;
        protected String m_strTextureName;
        protected ModelExTextureFile m_xFile;
        public override String Name { get { return m_strTextureName; } }
        public override String FileName { get { return (m_xFile != null) ? m_xFile.FileName : null; } }
        public ModelExTextureFile File { get { return m_xFile; } }
        public Texture Texture { get { return (m_xFile != null) ? m_xFile.Texture : null; } }

        public static implicit operator Texture(ModelExTextureInstance xInstance)
        {
            return xInstance.Texture;
        }

        public ModelExTextureInstance(ModelExSystem xSystem, String strFileName)
        {
            m_xSystem = xSystem;

            try
            {
                m_strTextureName = Path.GetFileName(strFileName);
            }
            catch { }

            if (xSystem == null)
            {
                throw new Exception("Texture instance must have a system");
            }

            if (m_strTextureName == null)
            {
                throw new Exception("Texture instance must have a file name");
            }
        }

        public void Load(String strFilePath)
        {
            try
            {
                m_xFile = ModelExTextureFile.Create(m_xSystem, strFilePath);
            }
            catch
            {
                m_xFile = null;
            }
        }

        public virtual void Load(ModelExTextureFile xTextureFile)
        {
            m_xFile = xTextureFile;
        }

        public override void Save(String strFilePath, ImageFormat eFormat)
        {
            if (m_xFile != null)
            {
                m_xFile.Save(strFilePath, eFormat);
            }
        }
    }
}