using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Text;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using BenLincoln.TheLostWorlds.CDTextures;
using VertexPNT = Microsoft.DirectX.Direct3D.CustomVertex.PositionNormalTextured;
using VertexPCT = Microsoft.DirectX.Direct3D.CustomVertex.PositionColoredTextured;

namespace ModelEx3
{
    class ModelExSRObject : ModelExObject
    {
        protected SR1File m_xSR1File;

        #region Render States

        protected override TextureOperation DefaultColorOperation
        {
            get
            {
                if (m_xSR1File == null ||
                    m_xSR1File.m_eFileType == FileType.Object)
                {
                    return TextureOperation.Modulate;
                }
                else
                {
                    return TextureOperation.Modulate4X;
                }
            }
        }
        protected override bool DefaultLighting
        {
            get
            {
                if (m_xSR1File == null ||
                    m_xSR1File.m_eFileType == FileType.Object)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
        }
        protected override ColorSource DefaultAmbientMaterialSource
        {
            get
            {
                if (m_xSR1File == null ||
                    m_xSR1File.m_eFileType == FileType.Object)
                {
                    return ColorSource.Material;
                }
                else
                {
                    return ColorSource.Color1;
                }
            }
        }
        protected override bool DefaultColorVertex
        {
            get
            {
                if (m_xSR1File == null ||
                    m_xSR1File.m_eFileType == FileType.Object)
                {
                    return false;
                }
                else
                {
                    return true;
                }
            }
        }

        #endregion

        public override String TextureArchives
        {
            get
            {
                switch (m_xSR1File.m_ePlatform)
                {
                    case Platform.PC:
                        return
                            "Soul Reaver Monolithic Files (*.big)|*.big|" +
                            "All Files (*.*)|*.*";
                    case Platform.PSX:
                        return
                            "Soul Reaver CRM Files (*.crm)|*.crm|" +
                            "Soul Reaver Archive Files (*.dat)|*.dat|" +
                            "All Files (*.*)|*.*";
                    case Platform.Dreamcast:
                        return
                            "Soul Reaver Archive Files (*.vq)|*.vq|" +
                            "All Files (*.*)|*.*";
                    default:
                        return null;
                }
            }
        }

        public override String TextureArchiveExt
        {
            get
            {
                switch (m_xSR1File.m_ePlatform)
                {
                    case Platform.PC:           return "big";
                    case Platform.PSX:          return "crm";
                    case Platform.Dreamcast:    return "vq";
                    default:                    return null;
                }
            }
        }

        public ModelExSRObject(String strFileName, ref Device xDevice)
            : base(strFileName, ref xDevice)
        {
        }

        protected override void Load()
        {
            if (!m_bInitialised)
            {
                #region Initialise

                // TextureGutterHelper <----- USE THIS!!!!!

                m_xSR1File = new SR1File(m_strFileName);
                m_strObjectName = m_xSR1File.m_strModelName.TrimEnd(new char[] { '_' });
                {
                    m_strObjectName.Trim();
                }

                for (int iModelID = 0; iModelID < m_xSR1File.m_usModelCount; iModelID++)
                {
                    m_xMeshContainer = new ModelExMeshContainer();
                    m_xMeshContainer.Name = m_strObjectName + "_" + iModelID.ToString();

                    // Choose the vertex format
                    VertexFormats eVertexFormat =
                        (m_xSR1File.m_eFileType == FileType.Object) ? VertexPNT.Format : VertexPCT.Format;

                    // Create the Mesh
                    m_xMeshContainer.Mesh = new Mesh(
                        (int)m_xSR1File.m_axModels[iModelID].PolygonCount,
                        (int)m_xSR1File.m_axModels[iModelID].IndexCount,
                        MeshFlags.Managed,
                        eVertexFormat,
                        m_xDevice
                    );

                    // Insert the vertices
                    if (eVertexFormat == VertexPNT.Format)
                    {
                        VertexPNT[] xVerticesPNT = new VertexPNT[m_xSR1File.m_axModels[iModelID].IndexCount];
                        for (UInt16 v = 0; v < m_xSR1File.m_axModels[iModelID].IndexCount; v++)
                        {
                            xVerticesPNT[v] = new VertexPNT(
                                100f * (float)m_xSR1File.m_axModels[iModelID].Vertices[v].worldPos.x,
                                100f * (float)m_xSR1File.m_axModels[iModelID].Vertices[v].worldPos.z,
                                100f * (float)m_xSR1File.m_axModels[iModelID].Vertices[v].worldPos.y,
                                (float)m_xSR1File.m_axModels[iModelID].Vertices[v].normal.x,
                                (float)m_xSR1File.m_axModels[iModelID].Vertices[v].normal.z,
                                (float)m_xSR1File.m_axModels[iModelID].Vertices[v].normal.y,
                                m_xSR1File.m_axModels[iModelID].Vertices[v].fU,
                                m_xSR1File.m_axModels[iModelID].Vertices[v].fV
                            );
                        }
                        m_xMeshContainer.Mesh.VertexBuffer.SetData(xVerticesPNT, 0, LockFlags.None);
                    }
                    else
                    {
                        VertexPCT[] xVerticesPCT = new VertexPCT[m_xSR1File.m_axModels[iModelID].IndexCount];
                        for (UInt16 v = 0; v < m_xSR1File.m_axModels[iModelID].IndexCount; v++)
                        {
                            xVerticesPCT[v] = new VertexPCT(
                                100f * (float)m_xSR1File.m_axModels[iModelID].Vertices[v].worldPos.x,
                                100f * (float)m_xSR1File.m_axModels[iModelID].Vertices[v].worldPos.z,
                                100f * (float)m_xSR1File.m_axModels[iModelID].Vertices[v].worldPos.y,
                                (int)m_xSR1File.m_axModels[iModelID].Vertices[v].colour,
                                m_xSR1File.m_axModels[iModelID].Vertices[v].fU,
                                m_xSR1File.m_axModels[iModelID].Vertices[v].fV
                            );
                        }
                        m_xMeshContainer.Mesh.VertexBuffer.SetData(xVerticesPCT, 0, LockFlags.None);
                    }

                    // Insert the indices
                    UInt16[] iIndices = new UInt16[m_xSR1File.m_axModels[iModelID].IndexCount];
                    for (UInt16 i = 0; i < m_xSR1File.m_axModels[iModelID].IndexCount; i++)
                    {
                        if (m_xSR1File.m_axModels[iModelID].Indices[i] % 3 == 0) iIndices[i] = m_xSR1File.m_axModels[iModelID].Indices[i];
                        if (m_xSR1File.m_axModels[iModelID].Indices[i] % 3 == 1) iIndices[i] = m_xSR1File.m_axModels[iModelID].Indices[i + 1];
                        if (m_xSR1File.m_axModels[iModelID].Indices[i] % 3 == 2) iIndices[i] = m_xSR1File.m_axModels[iModelID].Indices[i - 1];
                    }
                    m_xMeshContainer.Mesh.IndexBuffer.SetData(iIndices, 0, LockFlags.None);

                    // Load the materials
                    Material xMaterial = new Material();
                    m_xMeshContainer.Materials = new Material[m_xSR1File.m_axModels[iModelID].MaterialCount];
                    m_xMeshContainer.Textures = new Texture[m_xSR1File.m_axModels[iModelID].MaterialCount];
                    m_xMeshContainer.TextureNames = new String[m_xSR1File.m_axModels[iModelID].MaterialCount];
                    AttributeRange[] axAttributeRange = new AttributeRange[m_xSR1File.m_axModels[iModelID].MaterialCount];
                    for (UInt16 m = 0; m < m_xSR1File.m_axModels[iModelID].MaterialCount; m++)
                    {
                        if (m_xSR1File.m_eFileType == FileType.Object && m_xSR1File.m_axModels[iModelID].Materials[m].textureUsed)
                        {
                            xMaterial.Diffuse = Color.White;
                        }
                        else
                        {
                            xMaterial.Diffuse = Color.FromArgb(
                                (Int32)(m_xSR1File.m_axModels[iModelID].Materials[m].colour)
                            );
                        }
                        m_xMeshContainer.Materials[m] = xMaterial;
                        m_xMeshContainer.TextureNames[m] = m_xSR1File.m_axModels[iModelID].Materials[m].textureName;
                        axAttributeRange[m].AttributeId = m;
                    }

                    // Create the effects
                    m_xMeshContainer.Effects = new Effect[m_xSR1File.m_axModels[iModelID].MaterialCount];
                    m_xMeshContainer.DefaultEffects = new bool[m_xSR1File.m_axModels[iModelID].MaterialCount];
                    for (int e = 0; e < m_xSR1File.m_axModels[iModelID].MaterialCount; e++)
                    {
                        LoadDefaultEffect(e);
                    }

                    // Get the attributes
                    Int32[] axAttributeArray = m_xMeshContainer.Mesh.LockAttributeBufferArray(LockFlags.Discard);
                    for (UInt16 p = 0; p < m_xSR1File.m_axModels[iModelID].PolygonCount; p++)
                    {
                        axAttributeArray[p] = m_xSR1File.m_axModels[iModelID].Polygons[p].material.ID;
                    }
                    m_xMeshContainer.Mesh.UnlockAttributeBuffer(axAttributeArray);

                    // Weld any duplicate vertices
                    m_xMeshContainer.Mesh.WeldVertices(
                        WeldEpsilonsFlags.WeldPartialMatches,
                        new WeldEpsilons(),
                        null
                    );

                    // If it's an area, compute the normals
                    if (m_xSR1File.m_eFileType == FileType.Area)
                    {
                        m_xMeshContainer.Mesh = NormalizeMesh(m_xMeshContainer.Mesh);
                    }

                    m_xMeshContainers.Add(m_xMeshContainer);
                }

                if (m_xMeshContainers.Count > 0)
                {
                    m_xMeshContainer = m_xMeshContainers[0];
                }

                #endregion

                m_bInitialised = true;
            }
        }

        public override void LoadTexturesFromFile(String strFileName, int iModelIndex)
        {
            switch (m_xSR1File.m_ePlatform)
            {
                case Platform.PC:
                    LoadTexturesFromPCMonolithicFile(strFileName, iModelIndex);
                    break;
                case Platform.PSX:
                    LoadTexturesFromCRM(strFileName, iModelIndex);
                    break;
                case Platform.Dreamcast:
                    LoadTexturesFromDCMonolithicFile(strFileName, iModelIndex);
                    break;
            }
        }

        protected void LoadTexturesFromCRM(String strFileName, int iModelIndex)
        {
            ModelExMeshContainer xMeshContainer = m_xMeshContainers[iModelIndex];
            SR1Model xModel = m_xSR1File.m_axModels[iModelIndex];

            // Copy the polygon data for texturing
            SoulReaverPlaystationTextureFile.SoulReaverPlaystationPolygonTextureData[] xPolygons =
                new SoulReaverPlaystationTextureFile.SoulReaverPlaystationPolygonTextureData[m_xSR1File.m_axModels[iModelIndex].PolygonCount];

            int polyNum = 0;
            foreach (ExPolygon xPolygon in xModel.Polygons)
            {
                xPolygons[polyNum].paletteColumn = xPolygon.paletteColumn;
                xPolygons[polyNum].paletteRow = xPolygon.paletteRow;
                xPolygons[polyNum].u = new int[3];
                xPolygons[polyNum].v = new int[3];
                xPolygons[polyNum].u[0] = (int)xPolygon.v1.u;
                xPolygons[polyNum].u[1] = (int)xPolygon.v2.u;
                xPolygons[polyNum].u[2] = (int)xPolygon.v3.u;
                xPolygons[polyNum].v[0] = (int)xPolygon.v1.v;
                xPolygons[polyNum].v[1] = (int)xPolygon.v2.v;
                xPolygons[polyNum].v[2] = (int)xPolygon.v3.v;
                xPolygons[polyNum].textureID = xPolygon.material.textureID;
                polyNum++;
            }

            String strPathName = strFileName;
            SoulReaverPlaystationTextureFile xTextureFile = new SoulReaverPlaystationTextureFile(strPathName);
            xTextureFile.BuildTexturesFromPolygonData(xPolygons, false, true);
            for (int m = 0; m < xModel.Materials.Length; m++)
            {
                ExMaterial xMaterial = xModel.Materials[m];
                xMeshContainer.TextureNames[m] = xMaterial.textureName;
                if (xMaterial.textureUsed)
                {
                    xMeshContainer.Textures[m] = xTextureFile.GetTexture(m_xDevice, xMaterial.textureID);
                }
            }
        }

        public bool LoadTexturesFromPCMonolithicFile(string filePath, int iModelIndex)
        {
            ModelExMeshContainer xMeshContainer = m_xMeshContainers[iModelIndex];
            SR1Model xModel = m_xSR1File.m_axModels[iModelIndex];

            SoulReaverPCTextureFile xTextureFile = new SoulReaverPCTextureFile(filePath);
            for (int m = 0; m < xModel.Materials.Length; m++)
            {
                ExMaterial xMaterial = xModel.Materials[m];
                if (xMaterial.textureUsed)
                {
                    xMeshContainer.Textures[m] = xTextureFile.GetTexture(m_xDevice, xMaterial.textureID);
                }
            }
            return true;
        }

        public bool LoadTexturesFromDCMonolithicFile(string filePath, int iModelIndex)
        {
            ModelExMeshContainer xMeshContainer = m_xMeshContainers[iModelIndex];
            SR1Model xModel = m_xSR1File.m_axModels[iModelIndex];

            SoulReaverDreamcastTextureFile xTextureFile = new SoulReaverDreamcastTextureFile(filePath);
            for (int m = 0; m < xModel.Materials.Length; m++)
            {
                ExMaterial xMaterial = xModel.Materials[m];
                if (xMaterial.textureUsed)
                {
                    xMeshContainer.Textures[m] = xTextureFile.GetTexture(m_xDevice, xMaterial.textureID);
                }
            }
            return true;
        }
    }
}
